#pragma once

#include <algorithm>
#include <cstddef>

template <class T>
class Dyn_array
{
public:
    explicit Dyn_array(size_t initial_capacity = 8);

    ~Dyn_array();

    size_t size() const;

    T& operator[](size_t);
    const T& operator[](size_t) const;

    void push_back(T const&);
    void pop_back();

    Dyn_array(Dyn_array const&) = delete;
    Dyn_array& operator=(Dyn_array const&) = delete;

private:
    size_t capacity_;
    size_t size_;
    T* data_;
};

template <class T>
Dyn_array<T>::Dyn_array(size_t initial_capacity)
        : capacity_(initial_capacity)
        , size_(0)
        , data_(new T[capacity_])
{ }

template <class T>
Dyn_array<T>::~Dyn_array()
{
    delete [] data_;
}

template <class T>
size_t Dyn_array<T>::size() const
{
    return size_;
}

template <class T>
T& Dyn_array<T>::operator[](size_t ix)
{
    return data_[ix];
}

template <class T>
const T& Dyn_array<T>::operator[](size_t ix) const
{
    return data_[ix];
}

template <class T>
void Dyn_array<T>::push_back(T const& val)
{
    if (size_ == capacity_) {
        size_t new_cap = std::max(size_ + 1, 2 * capacity_);
        T* other_data = new T[new_cap];

        try {
            for (size_t i = 0; i < size_; ++i) {
                other_data[i] = data_[i];
            }
        } catch (...) {
            delete [] other_data;
            throw;
        }

        capacity_ = new_cap;
        std::swap(data_, other_data);

        delete [] other_data;
    }

    data_[size_++] = val;
}

template <class T>
void Dyn_array<T>::pop_back()
{
    data_[--size_] = T();
}

