#include "xor_list.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

struct xl_node
{
    int data;
    uintptr_t others;
};

static uintptr_t combine(struct xl_node* prev, struct xl_node* next)
{
    return (uintptr_t)prev ^ (uintptr_t)next;
}

static struct xl_node* extract(uintptr_t both, struct xl_node* other)
{
    return (struct xl_node*)(both ^ (uintptr_t)other);
}

void xl_init(struct xor_list* xl)
{
    xl->curr = NULL;
    xl->prev = NULL;
}

void xl_destroy(struct xor_list* xl)
{
    while (!xl_is_empty(xl)) {
        xl_remove(xl);
    }
}

bool xl_is_empty(struct xor_list* xl)
{
    return xl->curr == NULL;
}

void xl_forward(struct xor_list* xl)
{
    struct xl_node* next = extract(xl->curr->others, xl->prev);
    xl->prev = xl->curr;
    xl->curr = next;
}

void xl_back(struct xor_list* xl)
{
    struct xl_node* prevprev = extract(xl->prev->others, xl->curr);
    xl->curr = xl->prev;
    xl->prev = prevprev;
}

int* xl_data(struct xor_list* xl)
{
    return &xl->curr->data;
}

void xl_insert(struct xor_list* xl, int value)
{
    struct xl_node* new_node = malloc(sizeof(struct xl_node));
    if (new_node == NULL) {
        perror("xl_insert");
        exit(1);
    }

    new_node->data = value;
    new_node->others = combine(xl->prev, xl->curr);

    if (xl->curr == NULL) {
        xl->prev = xl->curr = new_node;
        return;
    } else if (xl->curr == xl->prev) {
        xl->prev = new_node;
        return;
    }

    struct xl_node* prevprev = extract(xl->prev->others, xl->curr);
    struct xl_node* next     = extract(xl->curr->others, xl->prev);

    xl->prev->others = combine(prevprev, new_node);
    xl->curr->others = combine(next, new_node);

    xl->prev = new_node;
}

void xl_remove(struct xor_list* xl)
{
    if (xl->prev == xl->curr) {
        free(xl->curr);
        xl->prev = xl->curr = NULL;
        return;
    }

    struct xl_node* victim   = xl->curr;
    struct xl_node* prevprev = extract(xl->prev->others, xl->curr);
    struct xl_node* next     = extract(xl->curr->others, xl->prev);
    struct xl_node* nextnext = extract(next->others, xl->curr);

    xl->curr = next;

    xl->prev->others = combine(prevprev, next);
    xl->curr->others = combine(xl->prev, nextnext);

    free(victim);
}

